#!/usr/bin/env python3

#
# NOSA HEADER START
#
# The contents of this file are subject to the terms of the NASA Open
# Source Agreement (NOSA), Version 1.3 only (the "Agreement").  You may
# not use this file except in compliance with the Agreement.
#
# You can obtain a copy of the agreement at
#   docs/NASA_Open_Source_Agreement_1.3.txt
# or
#   https://sscweb.gsfc.nasa.gov/WebServices/NASA_Open_Source_Agreement_1.3.txt.
#
# See the Agreement for the specific language governing permissions
# and limitations under the Agreement.
#
# When distributing Covered Code, include this NOSA HEADER in each
# file and include the Agreement file at
# docs/NASA_Open_Source_Agreement_1.3.txt.  If applicable, add the
# following below this NOSA HEADER, with the fields enclosed by
# brackets "[]" replaced with your own identifying information:
# Portions Copyright [yyyy] [name of copyright owner]
#
# NOSA HEADER END
#
# Copyright (c) 2020-2021 United States Government as represented by
# the National Aeronautics and Space Administration. No copyright is
# claimed in the United States under Title 17, U.S.Code. All Other
# Rights Reserved.
#

"""
Package for accessing the NASA's Satellite Situation Center (SSC) web 
services https://sscweb.gsfc.nasa.gov/WebServices/REST/.

Copyright &copy; 2020-2021 United States Government as represented by the
National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S.Code. All Other Rights Reserved.

Notes
-----
This library does not currently implement the /graphs and /conjunction
resources.  If there is interest in these, support could be added
in the future.

The core functionality is implemented in the `sscws.sscws` sub-module.
New users should start by viewing the `sscws.sscws` sub-module.
"""


__version__ = "0.1.12"

#
# XML schema namespace
#
NS = 'http://sscweb.gsfc.nasa.gov/schema'
#
# Namespace for use in xml.etree.ElementTree.find*
#
ET_NS = '{' + NS + '}'

